"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const EventEmitter = require("events");
const getStream = require("get-stream");
const is_1 = require("@sindresorhus/is");
const PCancelable = require("p-cancelable");
const merge_1 = require("./merge");
const errors_1 = require("./errors");
const normalize_arguments_1 = require("./normalize-arguments");
const request_as_event_emitter_1 = require("./request-as-event-emitter");
exports.isProxiedSymbol = Symbol('proxied');
function asPromise(options) {
    const proxy = new EventEmitter();
    const parseBody = (response) => {
        if (options.responseType === 'json') {
            response.body = JSON.parse(response.body);
        }
        else if (options.responseType === 'buffer') {
            response.body = Buffer.from(response.body);
        }
        else if (options.responseType !== 'text' && !is_1.default.falsy(options.responseType)) {
            throw new Error(`Failed to parse body of type '${options.responseType}'`);
        }
    };
    const promise = new PCancelable((resolve, reject, onCancel) => {
        const emitter = request_as_event_emitter_1.default(options);
        onCancel(emitter.abort);
        const emitError = async (error) => {
            try {
                for (const hook of options.hooks.beforeError) {
                    // eslint-disable-next-line no-await-in-loop
                    error = await hook(error);
                }
                reject(error);
            }
            catch (error2) {
                reject(error2);
            }
        };
        emitter.on('response', async (response) => {
            proxy.emit('response', response);
            const stream = is_1.default.null_(options.encoding) ? getStream.buffer(response) : getStream(response, { encoding: options.encoding });
            let data;
            try {
                data = await stream;
            }
            catch (error) {
                emitError(new errors_1.ReadError(error, options));
                return;
            }
            if (response.req && response.req.aborted) {
                // Canceled while downloading - will throw a CancelError or TimeoutError
                return;
            }
            const limitStatusCode = options.followRedirect ? 299 : 399;
            response.body = data;
            try {
                for (const [index, hook] of options.hooks.afterResponse.entries()) {
                    // eslint-disable-next-line no-await-in-loop
                    response = await hook(response, updatedOptions => {
                        updatedOptions = normalize_arguments_1.reNormalizeArguments(merge_1.mergeOptions(options, Object.assign({}, updatedOptions, { 
                            // @ts-ignore TS complaining that it's missing properties, which get merged
                            retry: {
                                calculateDelay: () => 0
                            }, throwHttpErrors: false, responseType: 'text', resolveBodyOnly: false })));
                        // Remove any further hooks for that request, because we we'll call them anyway.
                        // The loop continues. We don't want duplicates (asPromise recursion).
                        updatedOptions.hooks.afterResponse = options.hooks.afterResponse.slice(0, index);
                        return asPromise(updatedOptions);
                    });
                }
            }
            catch (error) {
                emitError(error);
                return;
            }
            const { statusCode } = response;
            if (response.body) {
                try {
                    parseBody(response);
                }
                catch (error) {
                    if (statusCode >= 200 && statusCode < 300) {
                        const parseError = new errors_1.ParseError(error, response, options);
                        emitError(parseError);
                        return;
                    }
                }
            }
            if (statusCode !== 304 && (statusCode < 200 || statusCode > limitStatusCode)) {
                const error = new errors_1.HTTPError(response, options);
                if (emitter.retry(error) === false) {
                    if (options.throwHttpErrors) {
                        emitError(error);
                        return;
                    }
                    resolve(options.resolveBodyOnly ? response.body : response);
                }
                return;
            }
            resolve(options.resolveBodyOnly ? response.body : response);
        });
        emitter.once('error', reject);
        [
            'request',
            'redirect',
            'uploadProgress',
            'downloadProgress'
        ].forEach(event => emitter.on(event, (...args) => {
            proxy.emit(event, ...args);
        }));
    });
    promise[exports.isProxiedSymbol] = true;
    promise.on = (name, fn) => {
        proxy.on(name, fn);
        return promise;
    };
    promise.json = () => {
        options.responseType = 'json';
        options.resolveBodyOnly = true;
        return promise;
    };
    promise.buffer = () => {
        options.responseType = 'buffer';
        options.resolveBodyOnly = true;
        return promise;
    };
    promise.text = () => {
        options.responseType = 'text';
        options.resolveBodyOnly = true;
        return promise;
    };
    return promise;
}
exports.default = asPromise;
//# sourceMappingURL=as-promise.js.map