"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const errors = require("./errors");
const deep_freeze_1 = require("./utils/deep-freeze");
const merge_1 = require("./merge");
const as_promise_1 = require("./as-promise");
const as_stream_1 = require("./as-stream");
const normalize_arguments_1 = require("./normalize-arguments");
const getPromiseOrStream = (options) => options.stream ? as_stream_1.default(options) : as_promise_1.default(options);
const aliases = [
    'get',
    'post',
    'put',
    'patch',
    'head',
    'delete'
];
const defaultHandler = (options, next) => next(options);
// `got.mergeInstances()` is deprecated
let hasShownDeprecation = false;
const create = (defaults) => {
    defaults = merge_1.default({}, defaults);
    normalize_arguments_1.preNormalizeArguments(defaults.options);
    defaults = Object.assign({ handlers: [defaultHandler], options: {} }, defaults, { mutableDefaults: Boolean(defaults.mutableDefaults) });
    // @ts-ignore Because the for loop handles it for us, as well as the other Object.defines
    const got = (url, options) => {
        const isStream = options && options.stream;
        let iteration = 0;
        const iterateHandlers = (newOptions) => {
            let nextPromise;
            const result = defaults.handlers[iteration++](newOptions, options => {
                const fn = iteration === defaults.handlers.length ? getPromiseOrStream : iterateHandlers;
                if (isStream) {
                    return fn(options);
                }
                // We need to remember the `next(options)` result.
                nextPromise = fn(options);
                return nextPromise;
            });
            // Proxy the properties from the next handler to this one
            if (!isStream && !Reflect.has(result, as_promise_1.isProxiedSymbol)) {
                for (const key of Object.keys(nextPromise)) {
                    Object.defineProperty(result, key, {
                        get: () => {
                            return nextPromise[key];
                        },
                        set: (value) => {
                            nextPromise[key] = value;
                        }
                    });
                }
                result.cancel = nextPromise.cancel;
                result[as_promise_1.isProxiedSymbol] = true;
            }
            return result;
        };
        try {
            return iterateHandlers(normalize_arguments_1.normalizeArguments(url, options, defaults));
        }
        catch (error) {
            if (isStream) {
                throw error;
            }
            else {
                // @ts-ignore It's an Error not a response, but TS thinks it's calling .resolve
                return Promise.reject(error);
            }
        }
    };
    got.create = create;
    got.extend = (...instancesOrOptions) => {
        const options = [defaults.options];
        const handlers = [...defaults.handlers];
        let mutableDefaults;
        for (const value of instancesOrOptions) {
            if (Reflect.has(value, 'defaults')) {
                options.push(value.defaults.options);
                handlers.push(...value.defaults.handlers.filter(handler => handler !== defaultHandler));
                mutableDefaults = value.defaults.mutableDefaults;
            }
            else {
                options.push(value);
                if (Reflect.has(value, 'handlers')) {
                    handlers.push(...value.handlers);
                }
                mutableDefaults = value.mutableDefaults;
            }
        }
        handlers.push(defaultHandler);
        return create({
            options: merge_1.mergeOptions(...options),
            handlers,
            mutableDefaults
        });
    };
    got.mergeInstances = (parent, ...instances) => {
        if (!hasShownDeprecation) {
            console.warn('`got.mergeInstances()` is deprecated. We support it solely for compatibility - it will be removed in Got 11. Use `instance.extend(...instances)` instead.');
            hasShownDeprecation = true;
        }
        return parent.extend(...instances);
    };
    // @ts-ignore The missing methods because the for-loop handles it for us
    got.stream = (url, options) => got(url, Object.assign({}, options, { stream: true }));
    for (const method of aliases) {
        got[method] = (url, options) => got(url, Object.assign({}, options, { method }));
        got.stream[method] = (url, options) => got.stream(url, Object.assign({}, options, { method }));
    }
    Object.assign(got, Object.assign({}, errors, { mergeOptions: merge_1.mergeOptions }));
    Object.defineProperty(got, 'defaults', {
        value: defaults.mutableDefaults ? defaults : deep_freeze_1.default(defaults),
        writable: defaults.mutableDefaults,
        configurable: defaults.mutableDefaults,
        enumerable: true
    });
    return got;
};
exports.default = create;
//# sourceMappingURL=create.js.map