"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const url_1 = require("url");
const cacheable_lookup_1 = require("cacheable-lookup");
const is_1 = require("@sindresorhus/is");
const lowercaseKeys = require("lowercase-keys");
const url_to_options_1 = require("./utils/url-to-options");
const validate_search_params_1 = require("./utils/validate-search-params");
const supports_brotli_1 = require("./utils/supports-brotli");
const merge_1 = require("./merge");
const known_hook_events_1 = require("./known-hook-events");
let hasShownDeprecation = false;
// `preNormalize` handles static options (e.g. headers).
// For example, when you create a custom instance and make a request
// with no static changes, they won't be normalized again.
//
// `normalize` operates on dynamic options - they cannot be saved.
// For example, `body` is every time different per request.
// When it's done normalizing the new options, it performs merge()
// on the prenormalized options and the normalized ones.
exports.preNormalizeArguments = (options, defaults) => {
    if (is_1.default.nullOrUndefined(options.headers)) {
        options.headers = {};
    }
    else {
        options.headers = lowercaseKeys(options.headers);
    }
    if (options.prefixUrl) {
        options.prefixUrl = options.prefixUrl.toString();
        if (!options.prefixUrl.toString().endsWith('/')) {
            options.prefixUrl += '/';
        }
    }
    if (is_1.default.nullOrUndefined(options.hooks)) {
        options.hooks = {};
    }
    else if (!is_1.default.object(options.hooks)) {
        throw new TypeError(`Parameter \`hooks\` must be an object, not ${is_1.default(options.hooks)}`);
    }
    for (const event of known_hook_events_1.default) {
        if (is_1.default.nullOrUndefined(options.hooks[event])) {
            if (defaults && defaults.hooks) {
                if (event in defaults.hooks) {
                    options.hooks[event] = defaults.hooks[event].slice();
                }
            }
            else {
                options.hooks[event] = [];
            }
        }
    }
    if (is_1.default.number(options.timeout)) {
        options.gotTimeout = { request: options.timeout };
    }
    else if (is_1.default.object(options.timeout)) {
        options.gotTimeout = options.timeout;
    }
    delete options.timeout;
    const { retry } = options;
    options.retry = {
        calculateDelay: retryObject => retryObject.computedValue,
        methods: new Set(),
        statusCodes: new Set(),
        errorCodes: new Set(),
        maxRetryAfter: undefined
    };
    if (is_1.default.nonEmptyObject(defaults) && retry !== false) {
        options.retry = Object.assign({}, defaults.retry);
    }
    if (retry !== false) {
        if (is_1.default.number(retry)) {
            options.retry.limit = retry;
        }
        else {
            // eslint-disable-next-line @typescript-eslint/no-object-literal-type-assertion
            options.retry = Object.assign({}, options.retry, retry);
        }
    }
    if (!options.retry.maxRetryAfter && options.gotTimeout) {
        options.retry.maxRetryAfter = Math.min(...[options.gotTimeout.request, options.gotTimeout.connect].filter(n => !is_1.default.nullOrUndefined(n)));
    }
    if (is_1.default.array(options.retry.methods)) {
        options.retry.methods = new Set(options.retry.methods.map(method => method.toUpperCase()));
    }
    if (is_1.default.array(options.retry.statusCodes)) {
        options.retry.statusCodes = new Set(options.retry.statusCodes);
    }
    if (is_1.default.array(options.retry.errorCodes)) {
        options.retry.errorCodes = new Set(options.retry.errorCodes);
    }
    if (options.dnsCache) {
        const cacheableLookup = new cacheable_lookup_1.default({ cacheAdapter: options.dnsCache });
        options.lookup = cacheableLookup.lookup;
        delete options.dnsCache;
    }
    return options;
};
exports.normalizeArguments = (url, options, defaults) => {
    let urlArgument;
    if (is_1.default.plainObject(url)) {
        options = Object.assign({}, url, options);
        urlArgument = options.url || {};
        delete options.url;
    }
    else {
        urlArgument = url;
    }
    if (defaults) {
        options = merge_1.default({}, defaults.options, options ? exports.preNormalizeArguments(options, defaults.options) : {});
    }
    else {
        options = merge_1.default({}, exports.preNormalizeArguments(options));
    }
    if (!is_1.default.string(urlArgument) && !is_1.default.object(urlArgument)) {
        throw new TypeError(`Parameter \`url\` must be a string or object, not ${is_1.default(urlArgument)}`);
    }
    let urlObj;
    if (is_1.default.string(urlArgument)) {
        if (options.prefixUrl && urlArgument.startsWith('/')) {
            throw new Error('`url` must not begin with a slash when using `prefixUrl`');
        }
        if (options.prefixUrl) {
            urlArgument = options.prefixUrl + urlArgument;
        }
        urlArgument = urlArgument.replace(/^unix:/, 'http://$&');
        urlObj = url_to_options_1.default(new url_1.URL(urlArgument));
    }
    else if (is_1.default.urlInstance(urlArgument)) {
        urlObj = url_to_options_1.default(urlArgument);
    }
    else if (options.prefixUrl) {
        urlObj = Object.assign({}, url_to_options_1.default(new url_1.URL(options.prefixUrl)), urlArgument);
    }
    else {
        urlObj = urlArgument;
    }
    // Override both null/undefined with default protocol
    options = merge_1.default({ path: '' }, urlObj, { protocol: urlObj.protocol || 'https:' }, options);
    for (const hook of options.hooks.init) {
        const isCalled = hook(options);
        if (is_1.default.promise(isCalled)) {
            throw new TypeError('The `init` hook must be a synchronous function');
        }
    }
    const { prefixUrl } = options;
    Object.defineProperty(options, 'prefixUrl', {
        set: () => {
            throw new Error('Failed to set prefixUrl. Options are normalized already.');
        },
        get: () => prefixUrl
    });
    let { searchParams } = options;
    delete options.searchParams;
    // TODO: Remove this before Got v11
    if (options.query) {
        if (!hasShownDeprecation) {
            console.warn('`options.query` is deprecated. We support it solely for compatibility - it will be removed in Got 11. Use `options.searchParams` instead.');
            hasShownDeprecation = true;
        }
        searchParams = options.query;
        delete options.query;
    }
    if (is_1.default.nonEmptyString(searchParams) || is_1.default.nonEmptyObject(searchParams) || (searchParams && searchParams instanceof url_1.URLSearchParams)) {
        if (!is_1.default.string(searchParams)) {
            if (!(searchParams instanceof url_1.URLSearchParams)) {
                validate_search_params_1.default(searchParams);
            }
            searchParams = (new url_1.URLSearchParams(searchParams)).toString();
        }
        options.path = `${options.path.split('?')[0]}?${searchParams}`;
    }
    if (options.hostname === 'unix') {
        const matches = /(.+?):(.+)/.exec(options.path);
        if (matches) {
            const [, socketPath, path] = matches;
            options = Object.assign({}, options, { socketPath,
                path, host: '' });
        }
    }
    const { headers } = options;
    for (const [key, value] of Object.entries(headers)) {
        if (is_1.default.nullOrUndefined(value)) {
            delete headers[key];
        }
    }
    if (options.decompress && is_1.default.undefined(headers['accept-encoding'])) {
        headers['accept-encoding'] = supports_brotli_1.default ? 'gzip, deflate, br' : 'gzip, deflate';
    }
    if (options.method) {
        options.method = options.method.toUpperCase();
    }
    return options;
};
exports.reNormalizeArguments = (options) => exports.normalizeArguments(url_1.format(options), options);
//# sourceMappingURL=normalize-arguments.js.map