"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const url_1 = require("url");
const util_1 = require("util");
const EventEmitter = require("events");
const http = require("http");
const https = require("https");
const CacheableRequest = require("cacheable-request");
const toReadableStream = require("to-readable-stream");
const is_1 = require("@sindresorhus/is");
const http_timer_1 = require("@szmarczak/http-timer");
const timed_out_1 = require("./utils/timed-out");
const get_body_size_1 = require("./utils/get-body-size");
const is_form_data_1 = require("./utils/is-form-data");
const calculate_retry_delay_1 = require("./calculate-retry-delay");
const get_response_1 = require("./get-response");
const progress_1 = require("./progress");
const errors_1 = require("./errors");
const url_to_options_1 = require("./utils/url-to-options");
const dynamic_require_1 = require("./utils/dynamic-require");
const URLGlobal = typeof URL === 'undefined' ? require('url').URL : URL;
const URLSearchParamsGlobal = typeof URLSearchParams === 'undefined' ? require('url').URLSearchParams : URLSearchParams;
const getMethodRedirectCodes = new Set([300, 301, 302, 303, 304, 305, 307, 308]);
const allMethodRedirectCodes = new Set([300, 303, 307, 308]);
const withoutBody = new Set(['GET', 'HEAD']);
exports.default = (options, input) => {
    const emitter = new EventEmitter();
    const redirects = [];
    let currentRequest;
    let requestUrl;
    let redirectString;
    let uploadBodySize;
    let retryCount = 0;
    let shouldAbort = false;
    const setCookie = options.cookieJar ? util_1.promisify(options.cookieJar.setCookie.bind(options.cookieJar)) : null;
    const getCookieString = options.cookieJar ? util_1.promisify(options.cookieJar.getCookieString.bind(options.cookieJar)) : null;
    const agents = is_1.default.object(options.agent) ? options.agent : null;
    const emitError = async (error) => {
        try {
            for (const hook of options.hooks.beforeError) {
                // eslint-disable-next-line no-await-in-loop
                error = await hook(error);
            }
            emitter.emit('error', error);
        }
        catch (error2) {
            emitter.emit('error', error2);
        }
    };
    const get = async (options) => {
        const currentUrl = redirectString || requestUrl;
        if (options.protocol !== 'http:' && options.protocol !== 'https:') {
            throw new errors_1.UnsupportedProtocolError(options);
        }
        decodeURI(currentUrl);
        let requestFn;
        if (is_1.default.function_(options.request)) {
            requestFn = options.request;
        }
        else {
            requestFn = options.protocol === 'https:' ? https.request : http.request;
        }
        if (agents) {
            const protocolName = options.protocol === 'https:' ? 'https' : 'http';
            options.agent = agents[protocolName] || options.agent;
        }
        /* istanbul ignore next: electron.net is broken */
        // No point in typing process.versions correctly, as
        // process.version.electron is used only once, right here.
        if (options.useElectronNet && process.versions.electron) {
            const electron = dynamic_require_1.default(module, 'electron'); // Trick webpack
            requestFn = electron.net.request || electron.remote.net.request;
        }
        if (options.cookieJar) {
            const cookieString = await getCookieString(currentUrl);
            if (is_1.default.nonEmptyString(cookieString)) {
                options.headers.cookie = cookieString;
            }
        }
        let timings;
        const handleResponse = async (response) => {
            try {
                /* istanbul ignore next: fixes https://github.com/electron/electron/blob/cbb460d47628a7a146adf4419ed48550a98b2923/lib/browser/api/net.js#L59-L65 */
                if (options.useElectronNet) {
                    response = new Proxy(response, {
                        get: (target, name) => {
                            if (name === 'trailers' || name === 'rawTrailers') {
                                return [];
                            }
                            const value = target[name];
                            return is_1.default.function_(value) ? value.bind(target) : value;
                        }
                    });
                }
                const { statusCode } = response;
                const typedResponse = response;
                typedResponse.statusMessage = typedResponse.statusMessage || http.STATUS_CODES[statusCode];
                typedResponse.url = currentUrl;
                typedResponse.requestUrl = requestUrl;
                typedResponse.retryCount = retryCount;
                typedResponse.timings = timings;
                typedResponse.redirectUrls = redirects;
                typedResponse.request = { options };
                typedResponse.isFromCache = typedResponse.fromCache || false;
                delete typedResponse.fromCache;
                const rawCookies = typedResponse.headers['set-cookie'];
                if (options.cookieJar && rawCookies) {
                    let promises = rawCookies.map((rawCookie) => setCookie(rawCookie, typedResponse.url));
                    if (options.ignoreInvalidCookies) {
                        promises = promises.map(p => p.catch(() => { }));
                    }
                    await Promise.all(promises);
                }
                if (options.followRedirect && 'location' in typedResponse.headers) {
                    if (allMethodRedirectCodes.has(statusCode) || (getMethodRedirectCodes.has(statusCode) && (options.method === 'GET' || options.method === 'HEAD'))) {
                        typedResponse.resume(); // We're being redirected, we don't care about the response.
                        if (statusCode === 303) {
                            // Server responded with "see other", indicating that the resource exists at another location,
                            // and the client should request it from that location via GET or HEAD.
                            options.method = 'GET';
                        }
                        if (redirects.length >= 10) {
                            throw new errors_1.MaxRedirectsError(typedResponse, options);
                        }
                        // Handles invalid URLs. See https://github.com/sindresorhus/got/issues/604
                        const redirectBuffer = Buffer.from(typedResponse.headers.location, 'binary').toString();
                        const redirectURL = new URLGlobal(redirectBuffer, currentUrl);
                        redirectString = redirectURL.toString();
                        redirects.push(redirectString);
                        const redirectOptions = Object.assign({}, options, { port: undefined, auth: undefined }, url_to_options_1.default(redirectURL));
                        for (const hook of options.hooks.beforeRedirect) {
                            // eslint-disable-next-line no-await-in-loop
                            await hook(redirectOptions, typedResponse);
                        }
                        emitter.emit('redirect', response, redirectOptions);
                        await get(redirectOptions);
                        return;
                    }
                }
                get_response_1.default(typedResponse, options, emitter);
            }
            catch (error) {
                emitError(error);
            }
        };
        const handleRequest = (request) => {
            if (shouldAbort) {
                request.abort();
                return;
            }
            currentRequest = request;
            request.on('error', error => {
                if (typeof request.aborted === 'number' || error.message === 'socket hang up') {
                    return;
                }
                if (error instanceof timed_out_1.TimeoutError) {
                    error = new errors_1.TimeoutError(error, timings, options);
                }
                else {
                    error = new errors_1.RequestError(error, options);
                }
                if (emitter.retry(error) === false) {
                    emitError(error);
                }
            });
            timings = http_timer_1.default(request);
            progress_1.uploadProgress(request, emitter, uploadBodySize);
            if (options.gotTimeout) {
                timed_out_1.default(request, options.gotTimeout, options);
            }
            emitter.emit('request', request);
            const uploadComplete = () => {
                request.emit('upload-complete');
            };
            try {
                if (is_1.default.nodeStream(options.body)) {
                    options.body.once('end', uploadComplete);
                    options.body.pipe(request);
                    options.body = undefined;
                }
                else if (options.body) {
                    request.end(options.body, uploadComplete);
                }
                else if (input && (options.method === 'POST' || options.method === 'PUT' || options.method === 'PATCH')) {
                    input.once('end', uploadComplete);
                    input.pipe(request);
                }
                else {
                    request.end(uploadComplete);
                }
            }
            catch (error) {
                emitError(new errors_1.RequestError(error, options));
            }
        };
        if (options.cache) {
            const cacheableRequest = new CacheableRequest(requestFn, options.cache);
            const cacheRequest = cacheableRequest(options, handleResponse);
            cacheRequest.once('error', error => {
                if (error instanceof CacheableRequest.RequestError) {
                    emitError(new errors_1.RequestError(error, options));
                }
                else {
                    emitError(new errors_1.CacheError(error, options));
                }
            });
            cacheRequest.once('request', handleRequest);
        }
        else {
            // Catches errors thrown by calling requestFn(...)
            try {
                // @ts-ignore TS complains that URLSearchParams is not the same as URLSearchParams
                handleRequest(requestFn(options, handleResponse));
            }
            catch (error) {
                emitError(new errors_1.RequestError(error, options));
            }
        }
    };
    emitter.retry = (error) => {
        let backoff;
        retryCount++;
        try {
            backoff = options.retry.calculateDelay({
                attemptCount: retryCount,
                retryOptions: options.retry,
                error,
                computedValue: calculate_retry_delay_1.default({
                    attemptCount: retryCount,
                    retryOptions: options.retry,
                    error,
                    computedValue: 0
                })
            });
        }
        catch (error2) {
            emitError(error2);
            return false;
        }
        if (backoff) {
            const retry = async (options) => {
                try {
                    for (const hook of options.hooks.beforeRetry) {
                        // eslint-disable-next-line no-await-in-loop
                        await hook(options, error, retryCount);
                    }
                    await get(options);
                }
                catch (error) {
                    emitError(error);
                }
            };
            setTimeout(retry, backoff, Object.assign({}, options, { forceRefresh: true }));
            return true;
        }
        return false;
    };
    emitter.abort = () => {
        if (currentRequest) {
            currentRequest.abort();
        }
        else {
            shouldAbort = true;
        }
    };
    setImmediate(async () => {
        try {
            for (const hook of options.hooks.beforeRequest) {
                // eslint-disable-next-line no-await-in-loop
                await hook(options);
            }
            // Serialize body
            const { body, headers } = options;
            const isForm = !is_1.default.nullOrUndefined(options.form);
            const isJSON = !is_1.default.nullOrUndefined(options.json);
            const isBody = !is_1.default.nullOrUndefined(body);
            if ((isBody || isForm || isJSON) && withoutBody.has(options.method)) {
                throw new TypeError(`The \`${options.method}\` method cannot be used with a body`);
            }
            if (isBody) {
                if (isForm || isJSON) {
                    throw new TypeError('The `body` option cannot be used with the `json` option or `form` option');
                }
                if (is_1.default.object(body) && is_form_data_1.default(body)) {
                    // Special case for https://github.com/form-data/form-data
                    headers['content-type'] = headers['content-type'] || `multipart/form-data; boundary=${body.getBoundary()}`;
                }
                else if (!is_1.default.nodeStream(body) && !is_1.default.string(body) && !is_1.default.buffer(body)) {
                    throw new TypeError('The `body` option must be a stream.Readable, string, Buffer, Object or Array');
                }
            }
            else if (isForm) {
                if (!is_1.default.object(options.form)) {
                    throw new TypeError('The `form` option must be an Object');
                }
                headers['content-type'] = headers['content-type'] || 'application/x-www-form-urlencoded';
                options.body = (new URLSearchParamsGlobal(options.form)).toString();
            }
            else if (isJSON) {
                headers['content-type'] = headers['content-type'] || 'application/json';
                options.body = JSON.stringify(options.json);
            }
            // Convert buffer to stream to receive upload progress events (#322)
            if (is_1.default.buffer(body)) {
                options.body = toReadableStream(body);
                uploadBodySize = body.length;
            }
            else {
                uploadBodySize = await get_body_size_1.default(options);
            }
            if (is_1.default.undefined(headers['content-length']) && is_1.default.undefined(headers['transfer-encoding'])) {
                if ((uploadBodySize > 0 || options.method === 'PUT') && !is_1.default.undefined(uploadBodySize)) {
                    headers['content-length'] = String(uploadBodySize);
                }
            }
            if (!options.stream && options.responseType === 'json' && is_1.default.undefined(headers.accept)) {
                options.headers.accept = 'application/json';
            }
            requestUrl = options.href || (new URLGlobal(options.path, url_1.format(options))).toString();
            await get(options);
        }
        catch (error) {
            emitError(error);
        }
    });
    return emitter;
};
//# sourceMappingURL=request-as-event-emitter.js.map